# Nut
# by Eresse <eresse@eresse.net>

# External Includes
require 'uri'
require 'cgi'
require 'typor'
require 'rxio'

# Internal Includes
require 'nut/codes'
require 'nut/request'
require 'nut/response'

# Nut Module
module Nut

	# Service Handler Module
	module Handler

		# Extend BinDelim I/O Filter
		extend RxIO::IOFilters::BinDelim

		# Set Message Delimiter
		msg_delim "\n"

		# Only handle HTTP/1.1 Requests
		SUPPORTED_VERSION = 'http/1.1'

		# On Join:
		# RxIO Service Interface Callback - Triggered each time a Client connects.
		# @param [Hash] client An RxIO Client Hash
		def self.on_join client
			# NoOp
		end

		# On Drop:
		# RxIO Service Interface Callback - Triggered each time a Client disconnects.
		# @param [Hash] client An RxIO Client Hash
		def self.on_drop client
			# NoOp
		end

		# Handle Message:
		# RxIO Service Interface Callback - Triggered each time a message is received from a Client.
		# @param [Hash] client An RxIO Client Hash
		# @param [String] msg The message received from the Client
		def self.handle_msg client, msg

			# Pass Full Message to Request Builder (including Message Delimiter)
			handle_req_dline client, msg + @msg_delim
		end

		# Sub-Process Input:
		# RxIO Service Interface Callback - Triggered each time a chunk of data is received from a Client.
		# Some Content-Length-based Requests may not end with a CRLF - this pulls in any left-over data directly from the RxIO Input Buffer.
		# @param [Hash] client Client Hash
		def self.subprocess_input client

			# Acquire Client Request Context Hash
			creq = client[:nut].try(:[], :req) || {}

			# Pull any left-overs from Input Buffer for Content-Length Requests
			handle_req_dline client, creq[:client][:ibuf].slice!(0, creq[:client][:ibuf].bytesize) if Request.is_content_len? creq
		end

		# Handle Request Data Line:
		# Ensures availability of a Client Request Context before passing Assembling & Processing.
		# @param [Hash] client An RxIO Client Hash
		# @param [String] msg The message received from the Client (including newline)
		def self.handle_req_dline client, msg

			# Ensure & Acquire Client Request Context
			creq = ensure_creq client

			# Assemble & Process
			assemble_process_req creq, msg unless creq[:body_complete]
		end

		# Ensure Client Request Hash:
		# Ensures existence of Client Request Hash inside Client Context.
		# @param [Hash] client An RxIO Client Hash
		def self.ensure_creq client

			# Ensure Client Request Structure is available
			client[:nut] ||= {}
			client[:nut][:req] ||= {

				# Request Lines (full request, including headers)
				lines: [],

				# Request Headers
				headz: {},

				# Full Request Line (including verb, URI and version)
				req_l: nil,

				# Full Request Body (just the body though, no headers)
				req_b: '',

				# Chunk Info (for Chunked Transfer-Encoding)
				chunk: {},

				# Client
				client: client
			}
		end

		# Assemble & Process Request:
		# Assembles and processes HTTP Requests from individual data lines.
		# @param [Hash] creq Client Request Context
		# @param [String] msg The message received from the Client (including newline)
		def self.assemble_process_req creq, msg

			# Register Request Line
			creq[:lines] << msg

			# Build Request Line
			Request.build_req_line creq, msg

			# Check Version
			raise "Unsupported HTTP Version [#{creq[:version]}]" unless creq[:version].downcase == SUPPORTED_VERSION

			# Build Headers
			Request.build_header creq, msg

			# Build Request Body
			Request.build_req_body creq, msg

			# Process Request when Complete
			process_req creq[:client]
		end

		# Process Request for Client:
		# Constructs a complete Request Hash from a Client Request Context and passes it to the Request Handler Module's _handle_req_ method.
		# @param [Hash] client An RxIO Client Hash
		def self.process_req client

			# Check Request Body Complete
			return unless client[:nut][:req][:body_complete]

			# Pull Request Hash & Complete
			creq = client[:nut].delete :req
			uri = URI.parse creq[:uri]
			request = {
				peer: client[:peer].dclone,
				server: client[:local],
				verb: creq[:verb].downcase.to_sym,
				uri: uri,
				body: creq[:req_b],
				ctype: creq[:ctype],
				params: Hash[*(CGI.parse(uri.query || '').inject([]) { |a, e| a + [e.first, (e.last.size == 1) ? e.last.first : e.last] })].sym_keys,
				headers: creq[:headz].dclone
			}

			# Pull all Request Details
			Request.extract_req_data request

			# Prepare Response
			response = { headers: {}, body: '' }

			# Acquire Request Handler
			rh = client[:serv].request_handler

			# Begin
			begin

				# Pass to Request Handler
				rh.handle request, response

			# Rescue (Catch Exceptions raised during request handling)
			rescue Exception => e

				# Handle Error
				(rh.respond_to?(:on_err) ? rh : self).on_err e, request, response
			end

			# Respond to Client
			serve_response client, response

			# Drop
			client[:drop] = true
		end

		# Serve Response:
		# Sends out an HTTP Response to the Client.
		# @param [Hash] response Response Hash
		def self.serve_response client, response

			# Send out Response String
			write client, Response.build_response(response)
		end

		# Default Error Handler:
		# Throws up a 500 Internal Server Error, with a description of the exception in the response body.
		# @param [Exception] ex An exception that was raised inside the handle method
		# @param [Hash] request Request Hash
		# @param [Hash] response Response Hash
		def self.on_err ex, request, response

			# Throw up a 500 Internal Server Error
			response[:code] = 500

			# Spit out Exception in Response Body
			response[:body] << '<html><head><title>Error</title></head><body>'
			response[:body] << '<h1>Internal Server Error</h1>'
			response[:body] << "<h2>#{ex}</h2>"
			ex.backtrace.each { |b| response[:body] << " -> #{b}<br/>" }
			response[:body] << '</body></html>'
		end
	end
end
