# Nut
# by Eresse <eresse@eresse.net>

# Nut Module
module Nut

	# Request Module
	module Request

		# Content-Type Regex
		CTYPE_REX = /^[^;]+/

		# Boundary Regex
		BOUNDARY_REX = /boundary=(.+)$/

		# Multipart Field Regex
		MULTIPART_FIELD_REX = /^([^=]+)=(.+)$/

		# Expect: 100-continue
		EXPECT_CONTINUE_REX = /100.continue/i

		# Pull Request Details:
		# Extracts any possible Form Data and extended Request details from a Request Hash.
		# @param [Hash] req Request Hash
		def self.extract_req_data req

			# Switch on Content-Type { Extract Form Data }
			case req[:ctype]
				when 'multipart/form-data'
					extract_multipart_form_data req
				when 'application/x-www-form-urlencoded'
					extract_urlencoded_form_data req
				else
					# NoOp
			end
		end

		# Extract Multipart Form Data:
		# Extracts Multipart Form Data from the Request Body and merges it into _req[:params]_.
		# @param [Hash] req Request Hash
		def self.extract_multipart_form_data req

			# Acquire Boundary
			req[:boundary] = BOUNDARY_REX.match(req[:headers][:content_type]).try(:[], 1).try :gsub, /[ \t]*$/, ''

			# Split Body into Parts
			req[:parts] = req[:body].split("\r\n--#{req[:boundary]}").reject { |p| p.empty? }

			# Remove last part ('--')
			req[:parts].pop

			# Collect Form Data from Parts
			req[:parts].each do |p|

				# Split Headers & Data
				headers, data = p.split "\r\n\r\n", 2

				# Generate Headers
				headers = build_gen_head headers.split("\r\n")

				# Extract additional information from headers
				finfo = Hash[*(headers[:content_disposition].split('; ').inject([]) do |a, f|
					m = MULTIPART_FIELD_REX.match(f)
					if m
						name = m.try :[], 1
						val = m.try :[], 2
						val = val.slice(1, val.size - 2) if val.start_with?('"') && val.end_with?('"')
						a + [name.downcase.to_sym, val]
					else
						a
					end
				end)]

				# Merge Request Params
				req[:params][finfo[:name].to_sym] = finfo[:filename] ? { filename: finfo[:filename], data: data } : data
			end
		end

		# Extract URL-Encoded Form Data:
		# Extracts URL-Encoded Form Data from the Request Body and merges it into _req[:params]_.
		# @param [Hash] req Request Hash
		def self.extract_urlencoded_form_data req

			# Split Body into Params & Merge
			req[:body]
				.split('&')
				.collect { |p| p.split '=', 2 }
				.each { |name, val| req[:params][name] = val }
		end

		# Build Request Line:
		# Parses *msg* as the Request Line unless already set, extracting HTTP Verb, URI & Version.
		# @param [Hash] creq Client Request Context
		# @param [String] msg A line of text to be processed
		def self.build_req_line creq, msg

			# Check Request Line
			return if creq[:req_l]

			# Set Request Line
			rl = msg.chomp
			creq[:req_l] = rl

			# Pull up Verb, URI, Version
			creq[:verb] = rl.slice 0, rl.index(' ')
			creq[:uri] = rl.slice rl.index(' ') + 1, rl.size
			creq[:version] = creq[:uri].slice! creq[:uri].rindex(' ') + 1, creq[:uri].size
			creq[:uri].chomp! ' '
		end

		# Build Header:
		# Extracts HTTP Headers from Request if available and not already set, returning true on success, false otherwise.
		# @param [Hash] creq Client Request Context
		# @param [String] msg A line of text to be processed
		def self.build_header creq, msg

			# Pull Headers if not already done
			if msg.chomp == '' && creq[:headz].empty?

				# Build Headers
				creq[:headz] = build_gen_head creq[:lines].dclone.slice(1, creq[:lines].size)

				# Acquire Clean Content-Type
				creq[:ctype] = CTYPE_REX.match(creq[:headz][:content_type]).try :[], 0

				# Respond to 100-continue Requests
				Nut::Handler.write creq[:client], "HTTP/1.1 100 Continue\r\n\r\n" if EXPECT_CONTINUE_REX =~ creq[:headz][:expect]
			end
		end

		# Build Generic Headers:
		# Constructs a structured Header Hash from an Array of text lines.
		# @param [Array] lines A bunch of lines containing header data
		# @return [Hash] A Hash representation of the Headers
		def self.build_gen_head lines

			# Prepare Header
			headz = {}

			# Pull up Header lines as { "Field-Name" => [val0, val1, ..., valX] }
			lines
				.inject([]) { |a, l| if /^[ \t]+/ =~ l then a.last << l.gsub(/^[ \t]+/, ' ') else a << l end; a }
				.collect { |l| l.chomp.split ': ', 2 }
				.select { |name, _val| name }
				.each { |name, val| headz[name] ||= []; headz[name] << val }

			# Concat Header fields into { field_name: 'field0,field1,...,fieldX' }
			Hash[*(headz.inject([]) { |a, h| a + [h.first.downcase.gsub('-', '_').to_sym, h.last.join(',')] })]
		end

		# Build Request Body:
		# Builds Request Body according to Request Headers (Content-Length / Transfer-Encoding).
		# @param [Hash] creq Client Request Context
		# @param [String] msg A line of text to be processed
		def self.build_req_body creq, msg

			# Check for Headers
			return if creq[:headz].empty?

			# Handle Chunked Transfer-Encoding
			build_body_chunked creq, msg

			# Handle Simple Body
			build_body_normal creq, msg

			# Handle No Body
			build_no_body creq
		end

		# Build Body - Chunked:
		# Assembles Chunks from *msg* (if Headers are complete) and assembles Request Body from them.
		# @param [Hash] creq Client Request Context
		# @param [String] msg A line of text to be processed
		def self.build_body_chunked creq, msg

			# Check for Chunked Transfer-Encoding
			return unless is_chunked? creq

			# Are we already reading a chunk?
			if creq[:chunk][:size]

				# Pull Chunk Data
				creq[:chunk][:data] ||= ''
				creq[:chunk][:data] << msg.chomp

				# Check Chunk Complete
				if creq[:chunk][:data].bytes.size >= creq[:chunk][:size]

					# Append to Body
					creq[:req_b] << creq[:chunk][:data].bytes.slice(0, creq[:chunk][:size]).pack('c*')

					# Reset Chunk
					creq[:chunk] = {}
				end
			else

				# Acquire next chunk size
				creq[:chunk][:size] = /[0-9]/.match(msg.chomp).try(:[], 0).try :to_i

				# Check Body Complete
				creq[:body_complete] = true if creq[:chunk][:size] == 0
			end
		end

		# Build Body - Normal (Content-Length):
		# Assembles Request Body from *msg* if Headers are complete, until Content-Length is reached.
		# @param [Hash] creq Client Request Context
		# @param [String] msg A line of text to be processed
		def self.build_body_normal creq, msg

			# Check for Content-Length
			return unless is_content_len? creq

			# Reject first empty line (except for multipart requests, in which the extra CRLF helps in splitting up request parts)
			if creq[:got_first_line] || (creq[:ctype] == 'multipart/form-data')

				# Assemble Request Body
				creq[:req_b] << msg

				# Check Body Complete
				if creq[:req_b].bytesize >= creq[:headz][:content_length].to_i
					creq[:body_complete] = true
					creq[:req_b] = creq[:req_b].bytes.slice(0, creq[:headz][:content_length].to_i).pack 'c*'
				end
			else
				creq[:got_first_line] = true
			end
		end

		# Build without Body:
		# Marks Request Body as Complete if no Body should be present (no content-length & no chunked encoding).
		# @param [Hash] creq Client Request Context
		def self.build_no_body creq

			# Check whether Body should be present
			return if expects_body? creq

			# Mark No Body
			creq[:body_complete] = true
		end

		# Is Content-Length?:
		# Determines whether request is bound by a content-length header.
		# @param [Hash] creq Client Request Context
		# @return [Boolean]
		def self.is_content_len? creq
			creq[:headz].try(:[], :content_length) && (!(creq[:headz].try(:[], :transfer_encoding)) || (creq[:headz].try(:[], :transfer_encoding) == 'identity'))
		end

		# Is Chunked?:
		# Determines whether request is chunked (as per transfer-encoding header).
		# @param [Hash] creq Client Request Context
		# @return [Boolean]
		def self.is_chunked? creq
			creq[:headz].try(:[], :transfer_encoding) && (creq[:headz].try(:[], :transfer_encoding) != 'identity')
		end

		# Expects Body?:
		# Determines whether request expects a body (according to headers).
		# @param [Hash] creq Client Request Context
		# @return [Boolean]
		def self.expects_body? creq
			is_content_len?(creq) || is_chunked?(creq)
		end
	end
end
